use crate::patterns::{Pattern, PixelColor};
use crate::Strip;
use serde::{Deserialize, Serialize};

/// # Colorwipe pattern
/// every pattern implement the iterator trait
///
/// This pattern fill the strip with a specific color, one by one.
///
/// ### Note
///
/// setting max_iteration to None lead the pattern to loop infinitely
///
#[derive(Serialize, Deserialize, Debug, Eq, Hash, PartialEq, Copy, Clone)]
pub struct ColorWipe<const N: usize> {
    #[serde(default)]
    pub(crate) current_iteration: usize,
    pub(crate) max_iteration: Option<usize>,
    pub(crate) color: PixelColor,
    pub(crate) background_color: Option<PixelColor>,
}

impl<const N: usize> Pattern for ColorWipe<N> {
    type Strip = Strip<N>;

    fn init(&mut self) -> Option<Self::Strip> {
        if let Some(color) = self.background_color {
            Some(Strip::<N>::new(color))
        } else {
            None
        }
    }

    fn is_last_iteration(&self) -> bool {
        match self.max_iteration {
            None => false,
            Some(max_iter) => {
                if self.current_iteration >= max_iter {
                    true
                } else {
                    false
                }
            }
        }
    }
}

impl<const N: usize> Iterator for ColorWipe<N> {
    type Item = Strip<N>;

    fn next(&mut self) -> Option<Self::Item> {
        if self.is_last_iteration() {
            return None;
        }

        let mut strip = Strip::default();
        if let Some(c) = self.background_color {
            strip.fill(c);
        }
        let iteration = self.current_iteration % N;

        for i in 0..iteration {
            strip[i] = Some(self.color);
        }
        self.current_iteration += 1;
        Some(strip)
    }
}
